-- Convenience Lua functions that can be used within rpm macros

-- Reads an rpm variable reference. Returns "%{" .. rpmvar.. "}" if rpmvar is
-- set to something, and nil otherwise. This is convenient in lua tests and to
-- distinguish between an unset variable and a variable set to something that
-- expands to "" at this point of the spec file.
local function ref(rpmvar)
  if not rpmvar or
    (rpm.expand("%{" .. rpmvar .. "}") == "%{" .. rpmvar .. "}") then
    return nil
  else
    return "%{" .. rpmvar .. "}"
  end
end

-- Expands <run> till there is no expansion left to do
local function expand(run)
  if not run then
    return nil
  end
  run = run:gsub("%%%%", "\029")
  local expanded = rpm.expand(run)
  while run ~= expanded do
    run = expanded
    expanded = rpm.expand(run)
  end
  expanded = expanded:gsub("\029", "%%%%")
  return expanded
end

-- Reads an rpm variable (reads the variable reference, and evaluates it).
-- Unlike a basic rpm.expand("{?" .. rpmvar .. "}"), returns nil if the
-- variable is unset, which is convenient in lua tests.
local function read(rpmvar)
  local ref = ref(rpmvar)
  if ref then
    return expand(ref)
  else
    return ref
  end
end

-- Sometimes it is convenient to distinguish between an unset variable and a
-- variable explicitly set to false. Unlike read, readbool returns a tristate,
-- true, false or nil
local function readbool(rpmvar)
  local value = read(rpmvar)
  if value then
    if (value:lower() == "false") then
      return false
    else
      return true
    end
  else
    return value
  end
end

-- Builds a list of un-commented and non-empty lines from a multiline rpm
-- variable. This allows using multiline expands for private %sourcelist-like
-- containers.
local function readlines(rpmvar)
  local lines = read(rpmvar)
  local L = {}
  if lines then
    for line in string.gmatch(lines,'[^\r\n]+') do
      if string.match(line, '^[^#]') then
        table.insert(L, line)
      end
    end
  end
  return L
end

-- Returns true if the macro that called this function had flag set
--   – for example, hasflag("z") would give the following results:
--     %foo -z bar → true
--     %foo -z     → true
--     %foo        → false
local function hasflag(flag)
  return (rpm.expand("%{-" .. flag .. "}") ~= "")
end

-- Returns the argument passed to flag in the macro that called this function
--  – for example, readflag("z") would give the following results:
--      %foo -z bar → bar
--      %foo        → nil
--      %foo -z ""  → empty string
--      %foo -z ''  → empty string
local function readflag(flag)
  if not hasflag(flag) then
    return nil
  else
    local a = rpm.expand("%{-" .. flag .. "*}")
    -- Handle "" and '' as empty strings
    if (a == '""') or (a == "''") then
      a = ''
    end
    return a
  end
end

-- %wordwrap core
local function wordwrap(text)
  text = expand(text .. "\n")
  text = text:gsub("%%%%", "\029")
  text = text:gsub("\t",               "  ")
  text = text:gsub("\r",               "\n")
  text = text:gsub(" +\n",             "\n")
  text = text:gsub("\n+\n",            "\n\n")
  text = text:gsub("^\n",              "")
  text = text:gsub("\n( *)[-*—][  ]+", "\n%1– ")
  output = ""
  for line in text:gmatch("[^\n]*\n") do
    local pos = 0
    local advance = ""
    for word in line:gmatch("%s*[^%s]*\n?") do
      local wl, bad = utf8.len(word)
      if not wl then
        -- Can not use warning in wordwrap since warning uses wordwrap
        rpm.expand([[
%{warn:Invalid UTF-8 sequence detected in:}
%{warn:]] .. word .. [[}
%{warn:It may produce unexpected results.}
]])
        wl = bad
      end
      if (pos == 0) then
        advance, n = word:gsub("^(%s*– ).*", "%1")
        if (n == 0) then
          advance  = word:gsub("^(%s*).*",   "%1")
        end
        advance = advance:gsub("– ", "  ")
        pos = pos + wl
      elseif  (pos + wl  < 81) or
             ((pos + wl == 81) and word:match("\n$")) then
        pos = pos + wl
      else
        word = advance .. word:gsub("^%s*", "")
        output = output .. "\n"
        pos = utf8.len(word)
      end
      output = output .. word
      if pos > 80 then
        pos = 0
        if not word:match("\n$") then
          output = output .. "\n"
        end
      end
    end
  end
  output = output:gsub("^\n*", "")
  output = output:gsub("\n*$", "\n")
  -- Escape %’s – we want macros that were preserved from wordwrap expansion to
  -- be preserved at the next stage too
  output = output:gsub("\029", "%%%%")
  return output
end

-- Reformats text and outputs it using a built-in rpm verb such as echo, warn
-- or error
local function message(verb, text)
  rpm.expand("%{" .. verb .. ":" .. wordwrap(text):gsub("\n*$", "") .. "}")
end

-- Writes some text to stdout
local function echo(text)
  message("echo", text)
end

-- Writes the list of z-suffixed rpmvars to the console, if set
local function echovars(rpmvars, z)
  for _, rpmvar in ipairs(rpmvars) do
    local suffixed = rpmvar .. z
    local header = string.sub("  " .. suffixed .. ":                                               ",1,24)
    local v = ref(suffixed)
    if v then
      echo(header .. v)
    end
  end
end

-- Writes some text as a warning
local function warning(text)
  message("warn", text)
end

-- Writes some text as an error
local function err(text)
  message("error", text)
end

-- Returns a table of <radical> = <value>, for all <radical>s in <radicals>,
-- with <value> = reference of <namespace>_<radical><suffix> if
-- <namespace>_<radical><suffix> is set. Radicals for which
-- <namespace>_<radical><suffix> is not set are omitted in results.
local function mread(radicals, namespace, suffix, resolve)
  local R = {}
  for _, radical in ipairs(radicals) do
    local name = radical
    if namespace and (namespace ~= "") then
      name = namespace .. "_" .. name
    end
    local ref = ref(name .. suffix)
    if ref then
      if resolve then
        ref = expand(ref)
      end
      R[radical] = ref
    end
  end
  return R
end

-- Sets a spec variable; echoes the result if verbose
local function set(rpmvar, value, verbose)
  if rpmvar then
    -- Empty the stack state
    while read(rpmvar) do
      rpm.undefine(rpmvar)
    end
    if (value == nil) then
      if ref(rpmvar) then
        if verbose then
          -- Already did it, white lie
          warning("Unsetting " .. rpmvar)
        end
      end
    else
      if (value == false) then
        value = "false"
      elseif (value == true) then
        value = "true"
      end
      if verbose then
        warning("Setting "  .. rpmvar .. " = " .. value)
      end
      rpm.define(rpmvar .. " %{expand:" .. value .. "}")
    end
  end
end

-- Unsets a spec variable if it is defined; echoes the result if verbose
local function unset(rpmvar, verbose)
  set(rpmvar, nil, verbose)
end

-- Sets a spec variable, if not already set; echoes the result if verbose
local function safeset(rpmvar, value, verbose)
  if not ref(rpmvar) then
    set(rpmvar, value, verbose)
  end
end

-- Sets the <namespace>_<radical><suffix> variable for all <radical>s in
-- <radicals>, to a <value> taken in the <stack> list of <radical> = <value>
-- tables. When <radical> exists in multiple tables, only its first <value> is
-- taken into account. When <radical> does not exist in any table, no
-- <namespace>_<radical><suffix> is set.
-- <stack> is usually constructed with multiple mread calls, taking
-- domain-specific fallback needs into consideration.
-- If <force> set via set (and unset variables with no matches), otherwise set
-- via safeset.
-- mset is a complex function; its specializations like alias, bialias, zalias
-- and set_current are sufficient for many needs.
local function mset(radicals, namespace, suffix, stack, verbose, force)
  for _, radical in ipairs(radicals) do
    local value = nil
    local found = false
    for _, values in ipairs(stack) do
      for k, v in pairs(values) do
        if k == radical then
          value = v
          found = true
          break
        end
      end
      if found then
        break
      end
    end
    local name = radical
    if namespace and (namespace ~= "") then
      name = namespace .. "_" .. name
    end
    if force then
      set(    name .. suffix, value, verbose)
    else
      safeset(name .. suffix, value, verbose)
    end
  end
end

-- Sets the <namespace>_<radical><suffix> variable for all <radical>s in
-- <radicals>, to a reference to <from_namespace>_<radical><from_suffix>,
-- if <from_namespace>_<radical><from_suffix> is set.
-- inherit_from is recursive: the next <from_suffix> is read in
-- <from_namespace>_<from_key><from_suffix>
local function inherit_from (radicals, namespace, suffix, from_namespace,
                             from_suffix, from_key, verbose)
  local seen  = {}
  local stack = {}
  local function icopy(z)
    if not z or seen[z] then
      return
    end
    seen[z] = true
    table.insert(stack, mread(radicals, from_namespace, z, false))
    if from_key and (from_key ~= "") then
      local new_key = from_key .. z
      if from_namespace and (from_namespace ~= "") then
        new_key = from_namespace .. "_" .. new_key
      end
      icopy(read(new_key))
    end
  end
  icopy(from_suffix)
  mset(radicals, namespace, suffix, stack, verbose, false, false)
end

-- Sets the <namespace>_<radical><suffix> variable for all <radical>s in
-- <radicals>, to a reference to <namespace>_<radical><from_suffix>,
-- if <namespace>_<radical><from_suffix> is set.
-- <from_suffix> is read in <namespace>_<from_key><suffix>
-- inherit is recursive, and will also process <from_key> at the next level
local function inherit(radicals, namespace, suffix, key, verbose)
  local k = key .. suffix
  if namespace and (namespace ~= "") then
    k = namespace .. "_" .. k
  end
  local from_suffix = read(k)
  inherit_from(radicals, namespace, suffix, namespace, from_suffix, key,
               verbose)
end

-- for each radical in radicals, safeset <to_namespace><radical><to_suffix>
-- to %{<from_namespace><radical><from_suffix>} if
-- <from_namespace><radical><from_suffix> is set to something
local function alias(radicals, from_namespace, from_suffix,
                               to_namespace,   to_suffix,   verbose)
  mset(radicals, to_namespace, to_suffix,
        {mread(radicals, from_namespace, from_suffix, false)},
        verbose, false)
end

-- Alias back and forth
local function bialias(radicals, namespace1, suffix1,
                                 namespace2, suffix2, verbose)
  alias(radicals, namespace1, suffix1, namespace2, suffix2, verbose)
  alias(radicals, namespace2, suffix2, namespace1, suffix1, verbose)
end

-- Aliases a list of rpm variables to the same variables suffixed with 0 (and
-- vice versa); echoes the result if verbose
local function zalias(rpmvars, verbose)
  bialias(rpmvars, "", "", "",  "0", verbose)
end

-- Takes a list of rpm variable roots and a suffix and alias __current_<rpmvar>
-- to <rpmvar><suffix> if it resolves to something not empty
local function set_current(rpmvars, suffix, verbose)
  mset(rpmvars, "__current", "",
       {mread(rpmvars, "", suffix, false)}, verbose, true)
end

-- Sets the usual verbosity variables in a generic way
local function set_verbose(verbose)
  if verbose then
    set(  "__current_verbose",       "-v",     verbose)
    unset("__current_quiet",                   verbose)
    unset("__current_shell_quiet",             verbose)
    unset("__current_shell_verbose",           verbose)
  else
    unset("__current_verbose",                 verbose)
    set(  "__current_quiet",         "-q",     verbose)
    set(  "__current_shell_quiet",   "set +x", verbose)
    set(  "__current_shell_verbose", "set -x", verbose)
  end
end

-- Returns an array of <namespace>_<radical>
local function qualify(radicals, namespace)
  local R = {}
  if (namespace ~= "") then
    namespace = namespace .. "_"
  end
  for _, radical in ipairs(radicals) do
    table.insert(R, namespace .. radical)
  end
  return R
end

-- Merges two lists, removing duplicates and ordering the result
-- The result is a list containing an ordered set of values
local function mergelists(list_of_lists)
  local S = {}
  local L = {}
  for   _, l in ipairs(list_of_lists) do
    for _, v in ipairs(l) do
      S[v] = true
    end
  end
  for k, _ in  pairs(S) do
    table.insert(L, k)
  end
  table.sort(L)
  return L
end

-- Returns a list of suffixes, for which <rpmvar><suffix> is set to something
local function suffixes(rpmvar)
  local S = {}
  for z=0,9999 do
    if ref(rpmvar .. z) then
      table.insert(S, tostring(z))
    end
  end
  -- rpm convention is to alias no suffix to zero suffix
  -- only add no suffix if zero suffix is different
  local value = read(rpmvar)
  if value and not (value == read(rpmvar .. "0")) then
    table.insert(S, "")
  end
  return S
end

-- Returns the list of suffixes, including the empty string, for which
-- <rpmvar><suffix> is set to a non empty value, for any rpmvar in rpmvars
local function all_suffixes(rpmvars)
  local SL = {}
  local S  = {}
  for _, rpmvar in ipairs(rpmvars) do
    table.insert(SL, suffixes(rpmvar))
  end
  S = mergelists(SL)
  table.sort(S,
             function(a,b) return (tonumber(a) or 0) < (tonumber(b) or 0) end)
  return S
end

local function suffix_new(set, verbose)
  local s = 0
  local used = false
  repeat
    z = tostring(s)
    used = false
    for k, _ in pairs(set) do
      used = used or ref(k .. z)
    end
    s = s + 1
  until (not used)
  for k, v in pairs(set) do
    safeset(k .. z, v, verbose)
  end
  return z
end

local function suffix_reuse(set, verbose)
  for s = 0, 9999 do
    z = tostring(s)
    local matches = true
    for k, v in pairs(set) do
      matches = matches and (read(k .. z) == v)
    end
    if matches then
      return z
    end
  end
  return suffix_new(set, verbose)
end

-- Takes a set[key] = <value> table and returns a suffix for which all the
-- <key><suffix> rpm variables have the corresponding <value> set.
-- If reuse attempts to find an existing suffix that matches, otherwise,
-- takes an available suffix and creates the variable set from scratch
local function suffix(set, reuse, verbose)
  if reuse then
    return suffix_reuse(set, verbose)
  else
    return suffix_new(set, verbose)
  end
end

-- %writevars core
local function writevars(macrofile, rpmvars)
  for _, rpmvar in ipairs(rpmvars) do
    print("sed -i 's\029" .. string.upper("@@" .. rpmvar .. "@@") ..
                   "\029" .. rpm.expand(  "%{" .. rpmvar .. "}" ) ..
                   "\029g' " .. macrofile .. "\n")
  end
end

-- Executes f(k) for k in range, returning a table of ks for which
-- f(k) = value. Used to compute a pre_range for pfloop for example
local function fmatch(f, range, value)
  local K = {}
  for _, k in ipairs(range) do
    if (f(k) == value) then
      table.insert(K, k)
    end
  end
  return K
end

-- Executes the <f> function in a loop
--   – the loop is controled by <index> (a string):
--     — if the <index> value given to floop is nil, loop over the whole
--       <range>. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute <f> for the specified <index> value only, if
--       <index> is part of <range>
--   – <f> is passed <index> then <otherargs> as arguments.
--   – <pre_range> index values are processed before the rest of <range>.
local function pfloop(f, index, pre_range, range, otherargs)
  local only  = {}
  local first = {}
  local last  = {}
  for _, i in ipairs(range) do
    if (index == nil) then
      local pre = false
      for _, p in ipairs(pre_range) do
        if (i == p) then
          pre = true
          break
        end
      end
      if pre then
        table.insert(first, i)
      else
        table.insert(last, i)
      end
    elseif (i == index) then
      table.insert(only, i)
      break
    end
  end
  for _, R in ipairs({only, first, last}) do
    for _, i in ipairs(R) do
      f(i, table.unpack(otherargs))
    end
  end
end

-- Executes the <f> function in a loop
--   – the loop is controled by <index> (a string):
--     — if the <index> value given to floop is nil, loop over the whole
--       <range>. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute <f> for the specified <index> value only, if
--       <index> is part of <range>
--   – <f> is passed <index> then <otherargs> as arguments.
--   – <pre_range> index values are processed before the rest of <range>.
local function floop(f, index, range, otherargs)
  pfloop(f, index, {}, range, otherargs)
end

-- Executes the <pkg> function in a loop, in a way compatible with package
-- header generation requirements
--   – the loop is controled by <index> (a string):
--     — if the <index> value given to floop is nil, loop over the whole
--       <range>. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute <pkg> for the specified <index> value only, if
--       <index> is part of <range>
--   – <pkg> is passed <index> then <otherargs> as arguments.
--   – <name>(<index>) must return the %{name} of the SRPM header that would
--     be generated if <pkg>(<index>, <otherargs>) is called before something
--     else generates the SRPM header.
--   – if %{source_name} is set, <name>(<index>) results will be compared to
--     %{source_name} for all valid <index> in the loop. <index> values for
--     which that comparison matches will be scheduled first
local function pkg_floop(pkg, index, otherargs, name, range)
  local source_range = {}
  local  source_name = read("source_name")
  if source_name then
    source_range = fmatch(name, range, source_name)
  end
  pfloop(pkg, index, source_range, range, otherargs)
end

-- A small helper to return <name>(<index>) for all <index> values in <range>, since
-- pkg_floop() will usually be used in conjunction with pkg_exec()
local function names(name, range)
  local N = {}
  for _, i in ipairs(range) do
    table.insert(N, name(i))
  end
  return N
end

-- A small helper to add function id to spec and console output, and keep track
-- of what is running when rpm barfs without providing any useful info
local function id(identifier)
  echo("  " .. identifier)
  print("#" .. identifier .. "\n")
end

--- Return true if a macro is defined and false otherwise.
local function is_defined(name)
  if macros then
    return macros[name] ~= nil
  end
  return rpm.expand("%{?" .. name .. ":1}") == "1"
end

--- Expand a macro. Return nil if the macro is undefined.
local function get_macro(name)
  if macros then
    return macros[name]
  end
  if not is_defined(name) then
    return nil
  end
  return rpm.expand("%{" .. name .. "}")
end

-- Get a flag. Pass true to has_arg if the flag accepts an argument.
-- For flags with arguments, return the flag argument or nil.
-- For flags without arguments, return a truthy value for defined or nil if the
-- flag was not passed.
local function get_flag(name, has_arg)
  if opt then
    return opt[name]
  end
  if not is_defined("-" .. name) then
    return nil
  end
  local ender = "}"
  if has_arg then
    ender = "*" .. ender
  end
  return rpm.expand("%{-" .. name .. ender)
end

-- deprecated graveyard
-- those can be removed as soon as a port of fonts-rpm-macros and go-rpm-macros
-- to the new API lands in koji

local function setcurrent(rpmvars, suffix, verbose)
  for _, rpmvar in ipairs(rpmvars) do
    local suffixed = rpmvar .. suffix
    if ref(suffixed) then
      set(  "current" .. rpmvar, "%{?" .. suffixed .. "}", verbose)
    else
      unset("current" .. rpmvar,                           verbose)
    end
  end
end

local function getsuffixed(rpmvar)
  local S = {}
  zalias({rpmvar}, false)
  for z=0,9999 do
    local value = read(rpmvar .. z)
    if value then
      S[tostring(z)] = value
    end
  end
  -- rpm convention is to alias no suffix to zero suffix
  -- only add no suffix if zero suffix is different
  local value = read(rpmvar)
  if value and (value ~= S["0"]) then
    S[""] = value
  end
  return S
end

local function getbestsuffix(rpmvar, value)
  local best         = nil
  local currentmatch = ""
  for z, setvalue in pairs(getsuffixed(rpmvar)) do
  if ((not best) or (string.len(setvalue) < string.len(currentmatch))) and
     (string.find(setvalue, "^" .. value)) then
      currentmatch = setvalue
      best         = z
    end
  end
  return best
end

return {
  ref            = ref,
  expand         = expand,
  read           = read,
  readbool       = readbool,
  readlines      = readlines,
  hasflag        = hasflag,
  readflag       = readflag,
  wordwrap       = wordwrap,
  echo           = echo,
  echovars       = echovars,
  warning        = warning,
  error          = err,
  mread          = mread,
  set            = set,
  unset          = unset,
  safeset        = safeset,
  mset           = mset,
  inherit_from   = inherit_from,
  inherit        = inherit,
  alias          = alias,
  bialias        = bialias,
  zalias         = zalias,
  set_current    = set_current,
  set_verbose    = set_verbose,
  qualify        = qualify,
  mergelists     = mergelists,
  suffixes       = suffixes,
  all_suffixes   = all_suffixes,
  suffix         = suffix,
  writevars      = writevars,
  floop          = floop,
  pkg_floop      = pkg_floop,
  names          = names,
  id             = id,
  is_defined     = is_defined,
  get_macro      = get_macro,
  get_flag       = get_flag,
  -- deprecated aliases
  explicitset    = set,
  explicitunset  = unset,
  getsuffixes    = suffixes,
  setcurrent     = setcurrent,
  getsuffixed    = getsuffixed,
  getbestsuffix  = getbestsuffix,
}
